// Copyright © 2022-2025 Gerald B. Cox
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

package beets

import (
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"log/slog"
	"net/http"
	"net/url"
	"os"
	"strings"
	"time"

	lb "codeberg.org/gbcox/beetbrainz/listenbrainz"
	"codeberg.org/gbcox/beetbrainz/types"
)

// Package-level variables for Beets configuration and HTTP client.
// These variables are typically initialized once from your main application's startup
// (e.g., from environment variables in your main.go or common.go).
var (
	beetsIP     string
	beetsPort   int
	fields      string       = "title,album,artist,mb_trackid,mb_albumid,mb_artistid,mb_releasegroupid,mb_workid,artists" // Added 'artists' to fields
	beetsClient *http.Client                                                                                              // Reusable HTTP client
)

// SetBeetsConfig allows other packages (like main/common) to set Beets connection details.
func SetBeetsConfig(ip string, port int) {
	beetsIP = ip
	beetsPort = port
	// Initialize beetsClient if not already done, or if config changes
	if beetsClient == nil {
		beetsClient = &http.Client{Timeout: 10 * time.Second}
	}
}

type BeetsData struct {
	Title          string   `json:"title"`
	Album          string   `json:"album"`
	Artist         string   `json:"artist"` // Main artist string, e.g., "Artist1 feat. Artist2"
	RecordingId    string   `json:"mb_trackid"`
	ReleaseId      string   `json:"mb_albumid"`
	ArtistId       string   `json:"mb_artistid"`
	ReleaseGroupId string   `json:"mb_releasegroupid"`
	WorkId         string   `json:"mb_workid"`
	Artists        []string `json:"artists"` // List of individual artists, e.g., ["Artist1", "Artist2"]
}

func (data *BeetsData) String() string {
	return fmt.Sprintf("%s - %s (%s)", data.Artist, data.Title, data.Album)
}

func GetMetadataForItem(item *types.MediaItem) (*lb.TrackMetadata, error) {
	// Initialize tm with basic info and the AdditionalInfo struct.
	// This ensures AdditionalInfo is always available.
	tm := lb.TrackMetadata{
		AdditionalInfo: &lb.AdditionalInfo{
			ListeningFrom: "Beetbrainz Media Server",
		},
		ArtistName:  item.Artist,
		ReleaseName: item.Album,
		TrackName:   item.Track,
	}

	var beetsData *BeetsData
	if os.Getenv("BEETS_IP") != "" {
		slog.Debug("Getting additional track metadata from Beets", "item", item.String())
		var err error
		beetsData, err = matchWithBeets(item)
		if err != nil {
			slog.Warn("Could not match with Beets", "error", err, "item", item.String())
			// We continue here even if Beets lookup fails, as original metadata might still be usable.
		}
	} else {
		slog.Warn("Beets IP not configured, skipping Beets lookup.")
	}

	if beetsData != nil {
		slog.Debug("Adding Beets data to ListenBrainz submission", "beets_item", beetsData.String())
		// Correctly assign MusicBrainz IDs to the nested AdditionalInfo struct
		tm.AdditionalInfo.ReleaseMbid = beetsData.ReleaseId
		// Use beetsData.ArtistId for the single ArtistMbid if available and primary
		if beetsData.ArtistId != "" {
			tm.AdditionalInfo.ArtistMbids = []string{beetsData.ArtistId}
		} else {
			// Fallback: collect IDs from beetsData.Artists if individual mbids aren't set
			// This might require additional data from Beets. For now, rely on main ArtistId.
			// If you need multiple artist MBIDs, BeetsData struct needs to be updated to capture `mb_artistids` (plural)
			// and then this logic expanded.
		}
		tm.AdditionalInfo.RecordingMbid = beetsData.RecordingId

		// Overwrite main TrackMetadata fields with Beets' more accurate data if available
		tm.ArtistName = beetsData.Artist
		tm.ReleaseName = beetsData.Album
		tm.TrackName = beetsData.Title
	} else {
		slog.Warn("No Beets data found for item, proceeding with original metadata (if available)", "item", item.String())
	}

	// Return the 'tm' variable, which has been correctly populated.
	return &tm, nil
}

func matchWithBeets(item *types.MediaItem) (*BeetsData, error) {
	if item.MusicBrainzTrackID != "" {
		// NEW ORDER: Attempt 0.1: Match by MusicBrainz Release Track ID (more specific, often sent by Emby)
		slog.Debug("Attempting match with Beets by MusicBrainz Release Track ID", "mb_track_id", item.MusicBrainzTrackID, "item", item.String())
		beetsResultsMBReleaseTrackID, err := getBeetsDataByID("mb_releasetrackid", item.MusicBrainzTrackID)
		if err == nil && len(beetsResultsMBReleaseTrackID) > 0 {
			slog.Debug("Found match with Beets by MusicBrainz Release Track ID", "webhook_item", item.String(), "beets_match", beetsResultsMBReleaseTrackID[0].String())
			return beetsResultsMBReleaseTrackID[0], nil
		}
		slog.Debug("Match by MusicBrainz Release Track ID failed or yielded no results, attempting match by MusicBrainz Track ID (fallback)", "mb_track_id", item.MusicBrainzTrackID, "item", item.String())

		// NEW ORDER: Attempt 0.2: Match by primary MusicBrainz Track ID (canonical song ID)
		beetsResultsMBID, err := getBeetsDataByID("mb_trackid", item.MusicBrainzTrackID)
		if err == nil && len(beetsResultsMBID) > 0 {
			slog.Debug("Found match with Beets by primary MusicBrainz Track ID", "webhook_item", item.String(), "beets_match", beetsResultsMBID[0].String())
			return beetsResultsMBID[0], nil
		}
		slog.Debug("Match by primary MusicBrainz Track ID also failed, attempting text-based match", "mb_track_id", item.MusicBrainzTrackID, "item", item.String())
	}

	// --- Attempt 1: Exact Match Query by title (original logic) ---
	slog.Debug("Attempting exact match with Beets by title", "item", item.String())
	beetsResultsExact, err := getBeetsDataByTitle(item.Track, true)
	if err != nil {
		slog.Error("Error during exact Beets query by title, falling back to fuzzy search", "error", err, "item", item.String())
		// Continue to fuzzy, don't return error yet
	} else if len(beetsResultsExact) > 0 {
		matchedBeetsData := findBestMatch(beetsResultsExact, item) // Your existing exact/partial matching logic
		if matchedBeetsData != nil {
			slog.Debug("Found exact match with Beets by title", "webhook_item", item.String(), "beets_match", matchedBeetsData.String())
			return matchedBeetsData, nil // Return immediately on success
		}
	}
	slog.Debug("Exact match by title failed, attempting fuzzy match by title", "item", item.String())

	// --- Attempt 2: Fuzzy Match Query by title (only if exact fails) ---
	beetsResultsFuzzy, err := getBeetsDataByTitle(item.Track, false)
	if err != nil {
		slog.Error("Error during fuzzy Beets query by title, no Beets data will be used", "error", err, "item", item.String())
		return nil, errors.New("failed to retrieve data from Beets after all attempts") // Indicate final failure
	}

	if len(beetsResultsFuzzy) == 0 {
		slog.Debug("No Beets data received for fuzzy item query by title", "item", item.String())
		return nil, errors.New("no Beets data received for query after fuzzy search by title")
	}

	matchedBeetsData := findBestMatch(beetsResultsFuzzy, item) // Re-use your matching logic
	if matchedBeetsData != nil {
		slog.Debug("Found fuzzy match with Beets by title", "webhook_item", item.String(), "beets_match", matchedBeetsData.String()) // Changed to Debug
		return matchedBeetsData, nil
	} else {
		// Only log all results if no suitable match was found for fuzzy
		var resultStrings []string
		for _, br := range beetsResultsFuzzy {
			resultStrings = append(resultStrings, br.String())
		}
		slog.Debug("No suitable Beets match found after fuzzy search by title", "webhook_item", item.String(), "beets_results_count", len(beetsResultsFuzzy), "beets_all_results", strings.Join(resultStrings, " || "))
		return nil, errors.New("no suitable match for item in Beets DB after all attempts")
	}
}

// NEW: getBeetsDataByID queries Beets using a specific MusicBrainz ID type.
func getBeetsDataByID(idType, id string) ([]*BeetsData, error) {
	beetsIP := os.Getenv("BEETS_IP")
	if beetsIP == "" {
		return nil, errors.New("BEETS_IP environment variable not set")
	}
	beetsPort := os.Getenv("BEETS_PORT")
	if beetsPort == "" {
		beetsPort = "8337" // Default port
	}

	// Beets query format for IDs is typically /item/query/mb_trackid:<ID>
	queryComponent := url.PathEscape(id)
	beetsURL := fmt.Sprintf("http://%s:%s/item/query/%s:%s", beetsIP, beetsPort, idType, queryComponent)

	return makeBeetsQuery(beetsURL) // Re-use the makeBeetsQuery helper
}

// Renamed: getBeetsData now getBeetsDataByTitle to reflect its purpose
func getBeetsDataByTitle(title string, exact bool) ([]*BeetsData, error) { // Added item for consistent logging
	beetsIP := os.Getenv("BEETS_IP")
	if beetsIP == "" {
		return nil, errors.New("BEETS_IP environment variable not set")
	}
	beetsPort := os.Getenv("BEETS_PORT")
	if beetsPort == "" {
		beetsPort = "8337" // Default port
	}

	if exact {
		// Attempt 1: Exact query with the original title (as received from webhook/item)
		beetsURL := buildBeetsURL("title", title, true, beetsIP, beetsPort) // New helper to build URL
		results, err := makeBeetsQuery(beetsURL)
		if err == nil && len(results) > 0 {
			slog.Debug("Exact query with original title returned results", "title", title, "count", len(results))
			return results, nil
		}
		slog.Debug("Exact query with original title failed or yielded no results, trying with normalized apostrophe", "title", title)

		// Attempt 2: If the original exact query failed, try again by replacing ASCII apostrophe with typographic one
		// This normalization is specific to Beets string handling for exact search.
		normalizedApostropheTitle := strings.ReplaceAll(title, "'", "’") // Replace ASCII with typographic
		if normalizedApostropheTitle != title {                          // Only query again if a change was made
			beetsURLNormalized := buildBeetsURL("title", normalizedApostropheTitle, true, beetsIP, beetsPort)
			results, errNormalized := makeBeetsQuery(beetsURLNormalized)
			if errNormalized == nil && len(results) > 0 {
				slog.Debug("Exact query with normalized apostrophe title returned results", "title", normalizedApostropheTitle, "count", len(results))
				return results, nil
			}
			slog.Debug("Exact query with normalized apostrophe title failed or yielded no results", "title", normalizedApostropheTitle)
		}

		// If both exact attempts fail, return no results.
		return nil, nil // Return no results, matchWithBeets will fall back to fuzzy
	} else {
		// This is the existing fuzzy logic from previous revisions.
		// It uses normalizeString principles but is specific to the Beets fuzzy query syntax.
		cleanedTitle := strings.ReplaceAll(title, "'", "")        // Remove ASCII apostrophes
		cleanedTitle = strings.ReplaceAll(cleanedTitle, "’", "")  // Remove typographic apostrophes
		cleanedTitle = strings.ReplaceAll(cleanedTitle, "/", " ") // Replace slashes with spaces

		// Split the title into words and URL-escape each word, then join with spaces.
		// This makes it behave more like "title:word1 word2 word3"
		words := strings.Fields(cleanedTitle) // Splits by whitespace
		var escapedWords []string
		for _, word := range words {
			escapedWords = append(escapedWords, url.PathEscape(word))
		}
		queryPart := strings.Join(escapedWords, " ")
		beetsURLFuzzy := buildBeetsURL("title", queryPart, false, beetsIP, beetsPort)
		return makeBeetsQuery(beetsURLFuzzy)
	}
}

// NEW: buildBeetsURL constructs the full URL for a Beets query.
func buildBeetsURL(queryType, queryValue string, isExact bool, beetsIP, beetsPort string) string {
	var finalQueryComponent string

	if isExact {
		// For exact queries, `queryValue` is the literal content.
		// We need to escape any slashes within the content itself for Beets.
		escapedContent := strings.ReplaceAll(queryValue, "/", "\\")
		finalQueryComponent = url.PathEscape(escapedContent)
	} else {
		// For fuzzy, `queryValue` is already prepared (words separated by spaces, URL-escaped)
		finalQueryComponent = queryValue
	}
	return fmt.Sprintf("http://%s:%s/item/query/%s:%s", beetsIP, beetsPort, queryType, finalQueryComponent)
}

// makeBeetsQuery is a new helper to encapsulate the actual HTTP request logic to Beets.
// It now takes the full Beets URL.
func makeBeetsQuery(beetsURL string) ([]*BeetsData, error) {
	client := &http.Client{
		Timeout: 10 * time.Second,
	}

	resp, err := client.Get(beetsURL)
	if err != nil {
		slog.Error("Beets request failed", "error", err, "url", beetsURL)
		return nil, fmt.Errorf("beets request failed: %w", err)
	}
	defer resp.Body.Close()

	if resp.StatusCode != http.StatusOK && resp.StatusCode != http.StatusAccepted {
		bodyBytes, _ := io.ReadAll(resp.Body)
		slog.Error("Beets API returned non-200 status code", "status_code", resp.StatusCode, "url_sent", beetsURL, "response_body", string(bodyBytes))
		return nil, fmt.Errorf("Beets API returned unexpected status code %d for '%s': %s", resp.StatusCode, beetsURL, string(bodyBytes))
	}

	var rj map[string][]*BeetsData
	err = json.NewDecoder(resp.Body).Decode(&rj)
	if err != nil {
		slog.Error("Failed to decode the response from Beets", "error", err, "url_sent", beetsURL)
		return nil, fmt.Errorf("failed to decode Beets response: %w", err)
	}
	return rj["results"], nil
}

// Revised normalizeString to handle apostrophes and dashes consistently.
func normalizeString(s string) string {
	s = strings.ReplaceAll(s, "'", "") // Remove ASCII apostrophe
	s = strings.ReplaceAll(s, "’", "") // Remove typographic apostrophe
	s = strings.ReplaceAll(s, "`", "") // Remove backtick
	s = strings.ReplaceAll(s, "-", "") // Remove hyphen
	s = strings.ReplaceAll(s, "–", "") // Remove en dash
	s = strings.ReplaceAll(s, "—", "") // Remove em dash

	// Consolidate multiple spaces into single, trim leading/trailing spaces
	s = strings.Join(strings.Fields(s), " ")
	return strings.ToLower(s) // Always return in lower case for case-insensitive comparison
}

// parseArtists splits an artist string into individual artist names.
// Handles common delimiters like ",", "feat.", "ft.", " and ", " & ".
func parseArtists(artistStr string) []string {
	if artistStr == "" {
		return nil
	}
	// Convert to lowercase first, then handle common delimiters.
	// Order of replacements matters for complex cases.
	temp := strings.ToLower(artistStr)

	// Replace common features/connectors with a standardized delimiter (e.g., "|")
	temp = strings.ReplaceAll(temp, "feat.", "|")
	temp = strings.ReplaceAll(temp, "ft.", "|")
	temp = strings.ReplaceAll(temp, " and ", "|")
	temp = strings.ReplaceAll(temp, " & ", "|")
	temp = strings.ReplaceAll(temp, ",", "|") // Replace commas last, after other multi-character delimiters

	// Split by the standardized delimiter
	parts := strings.Split(temp, "|")
	result := []string{}
	for _, p := range parts {
		trimmed := strings.TrimSpace(p)
		if trimmed != "" {
			result = append(result, trimmed)
		}
	}
	return result
}

// compareArtistLists checks if two lists of normalized artist names are essentially the same.
// Order doesn't matter. It normalizes inputs before comparison.
func compareArtistLists(listA, listB []string) bool {
	if len(listA) != len(listB) {
		return false
	}
	// Convert to maps for easy comparison, ignoring order, normalizing each element
	mapA := make(map[string]struct{})
	for _, s := range listA {
		mapA[normalizeString(s)] = struct{}{} // Normalize before putting into map
	}
	for _, s := range listB {
		if _, found := mapA[normalizeString(s)]; !found { // Normalize before lookup
			return false
		}
	}
	return true
}

// findBestMatch extracts the matching logic into a separate function,
// allowing matchWithBeets to decide on logging based on its return.
func findBestMatch(beetsResults []*BeetsData, refItem *types.MediaItem) *BeetsData {
	normalizedRefTrack := normalizeString(refItem.Track)
	normalizedRefAlbum := normalizeString(refItem.Album)

	// Parse incoming artist string into individual components once
	refArtistsComponents := parseArtists(refItem.Artist)

	for _, bd := range beetsResults {
		normalizedBdTitle := normalizeString(bd.Title)
		normalizedBdAlbum := normalizeString(bd.Album)

		if normalizedRefTrack == normalizedBdTitle && normalizedBdAlbum == normalizedRefAlbum { // Corrected comparison order for clarity
			// First, try to match using the individual artist components from BeetsData.Artists
			// BeetsData.Artists already provides individual artist names (e.g., ["Maroon 5", "Christina Aguilera"])
			if compareArtistLists(refArtistsComponents, bd.Artists) {
				slog.Debug("Found match by title/album and parsed artist lists", "item", refItem.String(), "beets_match_artist_list", bd.Artists)
				return bd
			}

			// Handle "Various Artists" specifically:
			// If the incoming item's artist is "Various Artists" (normalized),
			// and we have a title/album match, accept even if Beets has a specific artist.
			if normalizedRefArtist := normalizeString(refItem.Artist); normalizedRefArtist == normalizeString("Various Artists") {
				slog.Debug("Found specific artist in Beets for 'Various Artists' item with track/album match", "beets_artist", bd.Artist, "item", refItem.String())
				return bd
			}

			// Fallback for artist comparison: use the single artist string if list comparison fails
			// and it's not "Various Artists". Use an 'almostEqualSimple' for robustness.
			normalizedBdArtist := normalizeString(bd.Artist)                // This is "Maroon 5 feat. Christina Aguilera"
			normalizedRefArtist := normalizeString(refItem.Artist)          // This is "Maroon 5, Christina Aguilera"
			if almostEqualSimple(normalizedBdArtist, normalizedRefArtist) { // A simpler almostEqual for just the string
				slog.Debug("Found almostEqual match by title/album and single artist string", "item", refItem.String(), "beets_match_artist_string", bd.Artist)
				return bd
			}
		}
	}

	return nil // No suitable match found
}

// almostEqualSimple will be used for the single artist string fallback comparison.
// It performs a more forgiving comparison by removing common delimiters and whitespace
// before checking for equality or containment.
func almostEqualSimple(s1, s2 string) bool {
	// Helper to clean and normalize a string for comparison
	cleanAndNormalize := func(s string) string {
		s = strings.ToLower(s)
		s = strings.ReplaceAll(s, "feat.", "")
		s = strings.ReplaceAll(s, "ft.", "")
		s = strings.ReplaceAll(s, " and ", "")
		s = strings.ReplaceAll(s, " & ", "")
		s = strings.ReplaceAll(s, ",", "")
		s = strings.Join(strings.Fields(s), "") // Remove all spaces after cleaning delimiters for a dense comparison
		return s
	}

	cleanS1 := cleanAndNormalize(s1)
	cleanS2 := cleanAndNormalize(s2)

	return cleanS1 == cleanS2 || strings.Contains(cleanS1, cleanS2) || strings.Contains(cleanS2, cleanS1)
}
