package common

import (
	"crypto/rand"
	"encoding/hex"
	"fmt"
	"log/slog"
	"strings"
	"time"

	"codeberg.org/gbcox/beetbrainz/beets"
	env "codeberg.org/gbcox/beetbrainz/environ"
	lb "codeberg.org/gbcox/beetbrainz/listenbrainz"
	"codeberg.org/gbcox/beetbrainz/types"
)

// Request defines the common structure for all incoming webhook requests.
type Request struct {
	Event     string
	MediaType string
	User      string
	Item      *types.MediaItem
	Metadata  *lb.TrackMetadata
	Source    string // New, "mpris", "emby", "etc."
}

// Define generic event constants that media server handlers can use.
const (
	EventPlaybackStart = "start"
	EventPlaybackStop  = "stop"
)

// Define constants for ListenBrainz actions derived from events
const (
	LBACTION_PLAYING_NOW = "playing_now"
	LBACTION_SCROBBLE    = "scrobble"
	LBACTION_IGNORE      = "ignore"
)

// eventActionMap maps specific webhook event strings to ListenBrainz actions.
var eventActionMap = map[string]string{
	"playback.start":    LBACTION_PLAYING_NOW,
	"playback.unpause":  LBACTION_PLAYING_NOW,
	"playback.stop":     LBACTION_SCROBBLE,
	"playback.scrobble": LBACTION_SCROBBLE,
	"item.markplayed":   LBACTION_SCROBBLE,
	"media.play":        LBACTION_PLAYING_NOW,
	"media.resume":      LBACTION_PLAYING_NOW,
	"media.scrobble":    LBACTION_SCROBBLE,
	"media.stop":        LBACTION_SCROBBLE,
	"play":              LBACTION_PLAYING_NOW,
	"resume":            LBACTION_PLAYING_NOW,
	"scrobble":          LBACTION_SCROBBLE,
	"pause":             LBACTION_IGNORE,
	EventPlaybackStart:  LBACTION_PLAYING_NOW,
	EventPlaybackStop:   LBACTION_SCROBBLE,
}

// generateRequestID creates a random 128-bit hex string using the standard library.
func generateRequestID() string {
	b := make([]byte, 16)
	if _, err := rand.Read(b); err != nil {
		return "unknown-request-id"
	}
	return hex.EncodeToString(b)
}

// HandleRequest processes a generic webhook request from a media server.
func HandleRequest(req *Request) {
	requestID := generateRequestID()
	_ = requestID

	slog.Debug("HandleRequest invoked",
		"event", req.Event,
		"track", req.Item.Track,
		"album", req.Item.Album,
		"artist", req.Item.Artist,
		"mbid", req.Item.MusicBrainzTrackID,
		"media_type", req.MediaType,
		"source", req.Source,
		"session_id", req.Item.PlaySessionID,
		"timestamp", time.Now().Format(time.RFC3339Nano),
	)

	slog.Debug("Event received", "event", req.Event, "source", req.Source)

	apiToken := env.GetApiToken(req.User)
	if apiToken == "" {
		slog.Warn("No API token configured for user", "user", req.User, "event", req.Event, "item", req.Item.String())
		return
	}

	lbAction := eventActionMap[req.Event]
	if lbAction == "" {
		slog.Info("Unknown event received, ignoring", "event", req.Event, "media_type", req.MediaType, "user", req.User, "item", req.Item.String())
		return
	}

	if req.Source == "mpris" &&
		strings.EqualFold(req.Event, "playback.start") &&
		strings.EqualFold(req.MediaType, "audio") {

		slog.Debug("Submitting now playing (MPRIS)",
			"user", req.User,
			"track", req.Item.Track,
			"album", req.Item.Album,
		)

		tm, err := beets.GetMetadataForItem(req.Item)
		if err != nil {
			slog.Error("Failed to get Beets metadata for item", "error", err)
			return
		}

		err = lb.PlayingNow(apiToken, tm)
		if err != nil {
			slog.Error("Now playing submission failed", "error", err)
			return
		}

		slog.Debug("User is playing now (MPRIS)",
			"user", req.User,
			"artist", tm.ArtistName,
			"track", tm.TrackName,
			"album", tm.ReleaseName,
		)

		slog.Debug("✅ Exiting after MPRIS now playing — no scrobble will be submitted")
		return
	}

	if lbAction == LBACTION_IGNORE {
		slog.Info("Event mapped to ignore action, skipping", "event", req.Event, "media_type", req.MediaType, "user", req.User, "item", req.Item.String())
		return
	}

	if req.MediaType != "track" && req.MediaType != "audio" {
		slog.Info("Non-track/audio media type received, ignoring", "event", req.Event, "media_type", req.MediaType, "user", req.User, "item", req.Item.String())
		return
	}

	slog.Debug("Processing request",
		"event", req.Event,
		"lb_action", lbAction,
		"user", req.User,
		"item", req.Item.String(),
		"media_type", req.MediaType,
	)

	tm, err := beets.GetMetadataForItem(req.Item)
	if err != nil {
		slog.Error("Failed to get Beets metadata for item", "error", err, "item", req.Item.String(), "user", req.User)
		return
	}

	switch lbAction {
	case LBACTION_PLAYING_NOW:
		slog.Debug("Submitting scrobble to ListenBrainz",
			"user", req.User,
			"track", tm.TrackName,
			"album", tm.ReleaseName,
			"artist", tm.ArtistName,
		)
		err := lb.PlayingNow(apiToken, tm)
		if err != nil {
			slog.Error("Playing now request failed", "error", err, "item", req.Item.String(), "user", req.User)
		} else {
			slog.Debug("User is playing now",
				"user", req.User,
				"artist", tm.ArtistName,
				"track", tm.TrackName,
				"album", tm.ReleaseName,
			)
		}

	case LBACTION_SCROBBLE:
		sessionID := req.Item.PlaySessionID

		slog.Debug("Deduplication input check",
			"user", req.User,
			"trackID", req.Item.MusicBrainzTrackID,
			"sessionID", sessionID,
		)

		if sessionID != "" {
			slog.Debug("Checking for duplicate scrobble",
				"user", req.User,
				"track", req.Item.Track,
				"album", req.Item.Album,
				"mbid", req.Item.MusicBrainzTrackID,
				"session_id", req.Item.PlaySessionID,
				"source", req.Source,
			)

			if IsDuplicateScrobble(req.User, req.Item.MusicBrainzTrackID, sessionID, req.Source) {
				slog.Warn("Duplicate scrobble suppressed (session-based)",
					"user", req.User,
					"track", tm.TrackName,
					"artist", tm.ArtistName,
					"album", tm.ReleaseName,
					"mbid", req.Item.MusicBrainzTrackID,
					"session_id", sessionID,
					"source", req.Source,
				)
				return
			}
		} else {

			if IsRecentScrobble(req.User, req.Item.Track, req.Item.Album, req.Source) {
				slog.Warn("Duplicate scrobble suppressed (time-based fallback)",
					"user", req.User,
					"track", req.Item.Track,
					"album", req.Item.Album,
					"source", req.Source,
				)
				return
			}
		}

		slog.Debug("Scrobble accepted",
			"user", req.User,
			"track", tm.TrackName,
			"artist", tm.ArtistName,
			"album", tm.ReleaseName,
			"mbid", req.Item.MusicBrainzTrackID,
			"session_id", sessionID,
		)
		slog.Info(fmt.Sprintf("📡 Submitting scrobble: `%s`", tm.TrackName))

		err := lb.SubmitListen(apiToken, tm)
		if err != nil {
			slog.Error("Scrobble request failed", "error", err, "item", req.Item.String(), "user", req.User)
		} else {
			slog.Info(fmt.Sprintf("✅ User scrobbled: `%s`", tm.TrackName))
		}
	}
}
