// Copyright © 2022-2025 Gerald B. Cox
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

package emby

import (
	"bytes"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"log/slog"
	"net/http"
	"strings"

	"codeberg.org/gbcox/beetbrainz/common"
	"codeberg.org/gbcox/beetbrainz/types"
)

type EmbyExternalUrl struct {
	Name string `json:"Name"`
	Url  string `json:"Url"`
}

type EmbyItem struct {
	Type          string            `json:"Type"`
	Track         string            `json:"Name"`
	Album         string            `json:"Album"`
	AlbumArtist   string            `json:"AlbumArtist"`
	Artists       []string          `json:"Artists"`
	ExternalUrls  []EmbyExternalUrl `json:"ExternalUrls"`
	ProviderIds   map[string]string `json:"ProviderIds"`
	PlaySessionId string            `json:"PlaySessionId"` // This will be set manually from PlaybackInfo
}

func (item *EmbyItem) String() string {
	artist := item.AlbumArtist
	if len(item.Artists) > 0 && item.Artists[0] != "" {
		artist = strings.Join(item.Artists, ", ")
	}
	return fmt.Sprintf("%s - %s (%s)", artist, item.Track, item.Album)
}

func (item *EmbyItem) AsMediaItem() *types.MediaItem {
	mediaItem := &types.MediaItem{
		Album:         item.Album,
		Track:         item.Track,
		PlaySessionID: item.PlaySessionId,
	}

	if len(item.Artists) > 0 && item.Artists[0] != "" {
		mediaItem.Artist = strings.Join(item.Artists, ", ")
	} else {
		mediaItem.Artist = item.AlbumArtist
	}

	if item.ProviderIds != nil {
		if mbid, ok := item.ProviderIds["MusicBrainzArtist"]; ok {
			mediaItem.MusicBrainzArtistID = mbid
		}
		if mbid, ok := item.ProviderIds["MusicBrainzAlbum"]; ok {
			mediaItem.MusicBrainzAlbumID = mbid
		}
		if mbid, ok := item.ProviderIds["MusicBrainzTrack"]; ok {
			mediaItem.MusicBrainzTrackID = mbid
		}
		if mbid, ok := item.ProviderIds["MusicBrainzReleaseGroup"]; ok {
			mediaItem.MusicBrainzReleaseGroupID = mbid
		}
	} else if item.ExternalUrls != nil {
		for _, url := range item.ExternalUrls {
			switch url.Name {
			case "MusicBrainz Artist":
				if parts := strings.Split(url.Url, "/"); len(parts) > 0 {
					mediaItem.MusicBrainzArtistID = parts[len(parts)-1]
				}
			case "MusicBrainz Album":
				if parts := strings.Split(url.Url, "/"); len(parts) > 0 {
					mediaItem.MusicBrainzAlbumID = parts[len(parts)-1]
				}
			case "MusicBrainz Track":
				if parts := strings.Split(url.Url, "/"); len(parts) > 0 {
					mediaItem.MusicBrainzTrackID = parts[len(parts)-1]
				}
			case "MusicBrainz Release Group":
				if parts := strings.Split(url.Url, "/"); len(parts) > 0 {
					mediaItem.MusicBrainzReleaseGroupID = parts[len(parts)-1]
				}
			}
		}
	}

	return mediaItem
}

type EmbyAccount struct {
	Name string `json:"Name"`
}

type EmbyRequest struct {
	Event        string      `json:"Event"`
	Account      EmbyAccount `json:"User"`
	Item         EmbyItem    `json:"Item"`
	PlaybackInfo struct {
		PlaySessionId string `json:"PlaySessionId"`
	} `json:"PlaybackInfo"`
}

func HandleRequest(w http.ResponseWriter, r *http.Request) {
	if r.Method != "POST" {
		slog.Warn("Emby webhook received unsupported method", "method", r.Method, "expected_method", "POST")
		http.Error(w, "Method Not Allowed", http.StatusMethodNotAllowed)
		return
	}

	rq, err := parseRequest(r)
	if err != nil {
		slog.Error("Failed to parse Emby webhook request", "error", err)
		http.Error(w, fmt.Sprintf("Bad Request: %s", err.Error()), http.StatusBadRequest)
		return
	}

	common.HandleRequest(rq)
	w.WriteHeader(http.StatusOK)
}

func parseRequest(r *http.Request) (*common.Request, error) {
	var embyRequest EmbyRequest
	var err error

	bodyBytes, err := io.ReadAll(r.Body)
	if err != nil {
		return nil, fmt.Errorf("failed to read request body: %w", err)
	}
	r.Body = io.NopCloser(bytes.NewBuffer(bodyBytes))

	slog.Debug("Received Emby webhook raw body", "body", string(bodyBytes))

	contentType := r.Header.Get("Content-Type")

	if strings.Contains(contentType, "application/json") {
		err = json.NewDecoder(r.Body).Decode(&embyRequest)
		if err != nil {
			return nil, fmt.Errorf("failed to decode JSON from request body: %w", err)
		}
	} else if strings.Contains(contentType, "multipart/form-data") {
		err = r.ParseMultipartForm(16 << 20)
		if err != nil {
			return nil, fmt.Errorf("failed to parse multipart form: %w", err)
		}
		data := []byte(r.FormValue("data"))
		err = json.Unmarshal(data, &embyRequest)
		if err != nil {
			return nil, fmt.Errorf("failed to unmarshal JSON from 'data' form field: %w", err)
		}
	} else {
		return nil, fmt.Errorf("unsupported Content-Type: %s", contentType)
	}

	slog.Debug("Parsed Emby webhook request",
		"event", embyRequest.Event,
		"user_name", embyRequest.Account.Name,
		"item_type", embyRequest.Item.Type,
		"item_track", embyRequest.Item.Track,
		"item_album", embyRequest.Item.Album,
		"item_album_artist", embyRequest.Item.AlbumArtist,
		"item_track_artists", embyRequest.Item.Artists,
		"item_mbid_artist", embyRequest.Item.ProviderIds["MusicBrainzArtist"],
		"item_mbid_album", embyRequest.Item.ProviderIds["MusicBrainzAlbum"],
		"item_mbid_track", embyRequest.Item.ProviderIds["MusicBrainzTrack"],
	)

	if embyRequest.Event == "" {
		return nil, errors.New("Emby webhook payload missing 'Event' field")
	}
	if embyRequest.Item.Type != "Audio" {
		return nil, fmt.Errorf("Emby webhook 'Item.Type' is '%s', only 'Audio' (track) is supported", embyRequest.Item.Type)
	}
	if embyRequest.Item.Track == "" || embyRequest.Item.Album == "" {
		return nil, errors.New("Emby webhook payload missing required music metadata (Item.Name or Item.Album)")
	}
	if embyRequest.Account.Name == "" {
		return nil, errors.New("Emby webhook payload missing user 'User.Name'")
	}

	// ✅ Final missing link: assign session ID from PlaybackInfo
	embyRequest.Item.PlaySessionId = embyRequest.PlaybackInfo.PlaySessionId

	rq := &common.Request{
		Event:     embyRequest.Event,
		MediaType: strings.ToLower(embyRequest.Item.Type),
		User:      embyRequest.Account.Name,
		Item:      embyRequest.Item.AsMediaItem(),
		Source:    "emby", // ✅ Tower ID goes here
	}

	slog.Debug("Prepared common.Request for handling",
		"event", rq.Event,
		"media_type", rq.MediaType,
		"user", rq.User,
		"item_string_for_lookup", rq.Item.String(),
		"item_artist_for_lookup", rq.Item.Artist,
		"item_album_for_lookup", rq.Item.Album,
		"item_track_for_lookup", rq.Item.Track,
		"item_mbid_artist_for_lookup", rq.Item.MusicBrainzArtistID,
		"item_mbid_album_for_lookup", rq.Item.MusicBrainzAlbumID,
		"item_mbid_track_for_lookup", rq.Item.MusicBrainzTrackID,
	)

	slog.Debug("Sanity check: session ID propagation",
		"raw_session_id", embyRequest.Item.PlaySessionId,
		"media_item_session_id", rq.Item.PlaySessionID,
	)

	return rq, nil
}
