// Copyright © 2022-2025 Gerald B. Cox
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

package listenbrainz

import (
	"bytes"
	"encoding/json"
	"fmt"
	"io"
	"log/slog" // New: Import for structured logging
	"net/http"
	"time"
)

type AdditionalInfo struct {
	ListeningFrom string   `json:"listening_from,omitempty"`
	ReleaseMbid   string   `json:"release_mbid,omitempty"`
	ArtistMbids   []string `json:"artist_mbids,omitempty"`
	RecordingMbid string   `json:"recording_mbid,omitempty"`
	Tags          []string `json:"tags,omitempty"`
}

type TrackMetadata struct {
	AdditionalInfo *AdditionalInfo `json:"additional_info,omitempty"`
	ArtistName     string          `json:"artist_name"`
	TrackName      string          `json:"track_name"`
	ReleaseName    string          `json:"release_name"`
}

type ListenPayload struct {
	ListenedAt    int64          `json:"listened_at,omitempty"`
	TrackMetadata *TrackMetadata `json:"track_metadata"`
}

type Listen struct {
	ListenType string           `json:"listen_type"`
	Payload    []*ListenPayload `json:"payload"`
}

func PlayingNow(apiToken string, trackMetadata *TrackMetadata) error {
	l := Listen{
		ListenType: "playing_now",
		Payload: []*ListenPayload{
			{
				TrackMetadata: trackMetadata,
			},
		},
	}

	return submitRequest(apiToken, l)
}

func SubmitListen(apiToken string, trackMetadata *TrackMetadata) error {
	l := Listen{
		ListenType: "single",
		Payload: []*ListenPayload{
			{
				ListenedAt:    time.Now().Unix(),
				TrackMetadata: trackMetadata,
			},
		},
	}

	return submitRequest(apiToken, l)
}

func submitRequest(apiToken string, listen Listen) error {
	apiUrl := "https://api.listenbrainz.org/1/submit-listens"

	bdata, err := json.Marshal(listen)
	if err != nil {
		slog.Error("Failed to encode listen into JSON", "error", err, "listen_type", listen.ListenType)
		return err
	}

	for i := 1; i <= 5; i++ {
		r, err := http.NewRequest("POST", apiUrl, bytes.NewBuffer(bdata))
		if err != nil {
			slog.Error("Failed to create new HTTP request for ListenBrainz", "error", err, "attempt", i, "listen_type", listen.ListenType)
			return err
		}

		resp, err := doRequest(r, apiToken)
		if err == nil {
			if resp != nil {
				resp.Body.Close()
			}
			slog.Debug("ListenBrainz submission successful", "listen_type", listen.ListenType, "attempt", i) // Changed to Debug
			return nil
		}

		if resp != nil && resp.StatusCode < 500 {
			slog.Error("ListenBrainz submission failed with non-retriable status code", "status_code", resp.StatusCode, "error", err, "attempt", i, "listen_type", listen.ListenType)
			resp.Body.Close()
			return err
		}

		slog.Warn("ListenBrainz submission failed, retrying", "error", err, "attempt", i, "max_attempts", 5, "listen_type", listen.ListenType)
		time.Sleep(5 * time.Duration(i) * time.Second)
	}

	return fmt.Errorf("ListenBrainz submission failed after multiple retries for type %s", listen.ListenType)
}

func doRequest(r *http.Request, apiToken string) (*http.Response, error) {
	r.Header.Set("Content-Type", "application/json")
	r.Header.Set("Authorization", "Token "+apiToken)

	client := &http.Client{
		Timeout: 15 * time.Second,
	}

	resp, err := client.Do(r)
	if err != nil {
		return nil, fmt.Errorf("ListenBrainz request execution failed: %w", err)
	}

	if resp.StatusCode != http.StatusOK && resp.StatusCode != http.StatusAccepted {
		body, _ := io.ReadAll(resp.Body)
		slog.Error("ListenBrainz API returned unexpected status code", "status_code", resp.StatusCode, "response_body", string(body), "url", r.URL.String())
		return resp, fmt.Errorf("ListenBrainz API returned unexpected status code %d: %s", resp.StatusCode, string(body))
	}

	return resp, nil
}
