// Copyright © 2022-2025 Gerald B. Cox
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

package plex

import (
	"encoding/json"
	"errors"
	"fmt"
	"log/slog" // New: Import for structured logging
	"net/http"

	"codeberg.org/gbcox/beetbrainz/common"
	"codeberg.org/gbcox/beetbrainz/types"
)

type PlexItem struct {
	Type        string `json:"type"`
	Title       string `json:"title"`
	Parent      string `json:"parentTitle"`      // Album
	Grandparent string `json:"grandparentTitle"` // Artist
}

func (item *PlexItem) String() string {
	return fmt.Sprintf("%s - %s (%s)", item.Grandparent, item.Title, item.Parent)
}

func (item *PlexItem) AsMediaItem() *types.MediaItem {
	return &types.MediaItem{
		Artist: item.Grandparent,
		Album:  item.Parent,
		Track:  item.Title,
	}
}

type PlexAccount struct {
	Title string `json:"title"` // User's title/name in Plex webhook
}

type PlexRequest struct {
	Event   string      `json:"event"`
	Account PlexAccount `json:"Account"`  // Note: "Account" in JSON, not "account"
	Item    PlexItem    `json:"Metadata"` // Note: "Metadata" in JSON, not "item"
}

func HandleRequest(w http.ResponseWriter, r *http.Request) {
	if r.Method != "POST" {
		slog.Warn("Plex webhook received unsupported method", "method", r.Method, "expected_method", "POST")
		http.Error(w, "Method Not Allowed", http.StatusMethodNotAllowed)
		return
	}

	rq, err := parseRequest(r)
	if err != nil {
		slog.Error("Failed to parse Plex webhook request", "error", err)
		http.Error(w, fmt.Sprintf("Bad Request: %s", err.Error()), http.StatusBadRequest)
		return
	}

	common.HandleRequest(rq)

	w.WriteHeader(http.StatusOK)
}

func parseRequest(r *http.Request) (*common.Request, error) {
	err := r.ParseMultipartForm(1 << 20) // 1MB limit for memory
	if err != nil {
		return nil, fmt.Errorf("failed to parse multipart form data: %w", err)
	}

	payloadData := r.FormValue("payload")
	if payloadData == "" {
		return nil, errors.New("missing 'payload' form field in Plex webhook")
	}

	var plexRequest PlexRequest
	err = json.Unmarshal([]byte(payloadData), &plexRequest)
	if err != nil {
		return nil, fmt.Errorf("failed to unmarshal JSON from 'payload' field: %w", err)
	}

	if plexRequest.Event == "" {
		return nil, errors.New("Plex webhook payload missing 'event' field")
	}
	if plexRequest.Item.Type != "track" {
		return nil, fmt.Errorf("Plex webhook 'type' is '%s', only 'track' is supported", plexRequest.Item.Type)
	}
	if plexRequest.Item.Title == "" || plexRequest.Item.Parent == "" || plexRequest.Item.Grandparent == "" {
		return nil, errors.New("Plex webhook payload missing required music metadata (title, parentTitle, grandparentTitle)")
	}
	if plexRequest.Account.Title == "" {
		return nil, errors.New("Plex webhook payload missing user 'Account.title'")
	}

	rq := &common.Request{
		Event:     plexRequest.Event,
		MediaType: plexRequest.Item.Type,
		User:      plexRequest.Account.Title,
		Item:      plexRequest.Item.AsMediaItem(),
		Source:    "plex", // ✅ Tower ID goes here
	}

	return rq, nil
}
