// Copyright © 2022-2025 Gerald B. Cox
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

package tautulli

import (
	"encoding/json"
	"errors"
	"fmt"
	"log/slog" // New: Import for structured logging
	"net/http"

	"codeberg.org/gbcox/beetbrainz/common"
	"codeberg.org/gbcox/beetbrainz/types"
)

type TautulliRequest struct {
	Action      string `json:"action"`       // e.g., "play", "stop", "scrobble"
	UserName    string `json:"user_name"`    // User's name
	ArtistName  string `json:"artist_name"`  // Original artist name from Plex/Tautulli
	AlbumName   string `json:"album_name"`   // Original album name
	TrackName   string `json:"track_name"`   // Original track name
	TrackArtist string `json:"track_artist"` // Specific track artist (can differ from AlbumArtist)
	MediaType   string `json:"media_type"`   // e.g., "track"
}

func (r *TautulliRequest) String() string {
	return fmt.Sprintf("%s - %s (%s)", r.ArtistName, r.TrackName, r.AlbumName)
}

func (r *TautulliRequest) AsMediaItem() *types.MediaItem {
	artist := r.TrackArtist
	if artist == "" {
		artist = r.ArtistName
	}
	return &types.MediaItem{
		Artist: artist,
		Album:  r.AlbumName,
		Track:  r.TrackName,
	}
}

func HandleRequest(w http.ResponseWriter, r *http.Request) {
	if r.Method != "POST" {
		slog.Warn("Tautulli webhook received unsupported method", "method", r.Method, "expected_method", "POST")
		http.Error(w, "Method Not Allowed", http.StatusMethodNotAllowed)
		return
	}

	rq, err := parseRequest(r)
	if err != nil {
		slog.Error("Failed to parse Tautulli webhook request", "error", err)
		http.Error(w, fmt.Sprintf("Bad Request: %s", err.Error()), http.StatusBadRequest)
		return
	}

	common.HandleRequest(rq)

	w.WriteHeader(http.StatusOK)
}

func parseRequest(r *http.Request) (*common.Request, error) {
	var tautulliRequest TautulliRequest
	var err error

	err = json.NewDecoder(r.Body).Decode(&tautulliRequest)
	if err != nil {
		return nil, fmt.Errorf("failed to decode JSON from request body: %w", err)
	}

	if tautulliRequest.Action == "" {
		return nil, errors.New("Tautulli webhook payload missing 'action' field")
	}
	if tautulliRequest.MediaType != "track" && tautulliRequest.MediaType != "episode" {
		return nil, fmt.Errorf("Tautulli webhook 'media_type' is '%s', only 'track' is supported (or 'episode')", tautulliRequest.MediaType)
	}
	if tautulliRequest.TrackName == "" || tautulliRequest.AlbumName == "" {
		return nil, errors.New("Tautulli webhook payload missing required music metadata (track_name, album_name)")
	}
	if tautulliRequest.UserName == "" {
		return nil, errors.New("Tautulli webhook payload missing 'user_name'")
	}

	rq := &common.Request{
		Event:     tautulliRequest.Action,
		MediaType: tautulliRequest.MediaType,
		User:      tautulliRequest.UserName,
		Item:      tautulliRequest.AsMediaItem(),
		Source:    "tautulli", // ✅ Tower ID goes here
	}

	return rq, nil
}
