// Copyright © 2022-2025 Gerald B. Cox
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

package main

import (
	"context"
	"flag" // New import for command-line flags
	"fmt"
	"log/slog"
	"net"
	"net/http"
	"os"
	"os/signal"
	"strconv"
	"strings"
	"syscall"
	"time"

	"codeberg.org/gbcox/beetbrainz/emby"
	"codeberg.org/gbcox/beetbrainz/jellyfin"
	"codeberg.org/gbcox/beetbrainz/mpd"
	"codeberg.org/gbcox/beetbrainz/mpris"
	"codeberg.org/gbcox/beetbrainz/plex"
	"codeberg.org/gbcox/beetbrainz/tautulli"
)

// These variables will be set at compile time by the Go linker (-ldflags).
// They hold information about the build version and Git commit.
// Their names have been adjusted to avoid conflicts with automatic ldflags
// injection by RPM's %gobuild macro.
var (
	AppVersionFull  = "dev"  // The application version (e.g., 1.0.0), overridden by ldflags
	CommitFull      = "none" // The full Git commit hash, overridden by ldflags
	ShortCommitFull = "none" // The short Git commit hash, overridden by ldflags
)

func main() {
	// --- Command-line flag for version ---
	versionFlag := flag.Bool("version", false, "Print version information and exit")
	flag.Parse() // Parse the command-line arguments

	if *versionFlag {
		fmt.Printf("Beetbrainz Version: %s\n", AppVersionFull)                  // Use AppVersionFull
		fmt.Printf("Git Commit: %s (short: %s)\n", CommitFull, ShortCommitFull) // Use CommitFull, ShortCommitFull
		os.Exit(0)                                                              // Exit after printing version
	}

	// --- Configure Slog ---
	// Determine log level from environment variable, default to INFO if not set or invalid.
	var logLevel slog.Level
	switch strings.ToLower(os.Getenv("BEETBRAINZ_LOG_LEVEL")) {
	case "debug":
		logLevel = slog.LevelDebug
	case "info":
		logLevel = slog.LevelInfo
	case "warn":
		logLevel = slog.LevelWarn
	case "error":
		logLevel = slog.LevelError
	default:
		logLevel = slog.LevelInfo // Default to INFO if not specified
	}

	// Create a JSON handler with the determined log level.
	// Logs will be written to os.Stdout, making them easily parsable by log aggregation systems.
	handler := slog.NewJSONHandler(os.Stdout, &slog.HandlerOptions{
		Level: logLevel,
	})
	slog.SetDefault(slog.New(handler)) // Set the default logger with the configured handler

	// Log the configured log level.
	slog.Info("🎚️ Logging level set", "level", logLevel.String())

	// Log application version and build details immediately after setting up logging
	slog.Info("🧠 Starting Beetbrainz",
		"version", AppVersionFull,
		"commit", CommitFull,
		"shortCommit", ShortCommitFull)

	// --- Validate BEETBRAINZ_IP and BEETBRAINZ_PORT ---
	addr := os.Getenv("BEETBRAINZ_IP")
	if addr == "" {
		addr = "0.0.0.0" // Default to listen on all interfaces
	}

	portStr := os.Getenv("BEETBRAINZ_PORT")
	if portStr == "" {
		portStr = "5000" // Default port
	}
	port, err := strconv.Atoi(portStr)
	if err != nil {
		slog.Error("Invalid BEETBRAINZ_PORT. Must be a number.",
			"value", portStr,
			"error", err,
		)
		os.Exit(1)
	}
	if port < 1 || port > 65535 {
		slog.Error("BEETBRAINZ_PORT is out of valid range (1–65535).",
			"value", port,
		)
		os.Exit(1)
	}
	address := fmt.Sprintf("%s:%d", addr, port)

	// --- Validate BEETS_IP and BEETS_PORT (if Beets integration is enabled) ---
	beetsIP := os.Getenv("BEETS_IP")
	beetsPortStr := os.Getenv("BEETS_PORT")

	if beetsIP != "" { // Only validate Beets settings if BEETS_IP is provided
		if beetsPortStr == "" {
			beetsPortStr = "8337" // Default Beets port
		}
		beetsPort, err := strconv.Atoi(beetsPortStr)
		if err != nil {
			slog.Error("Invalid BEETS_PORT. Must be a number.",
				"value", beetsPortStr,
				"error", err,
			)
			os.Exit(1)
		}
		if beetsPort < 1 || beetsPort > 65535 {
			slog.Error("BEETS_PORT is out of valid range (1–65535).",
				"value", beetsPort,
			)
			os.Exit(1)
		}
		slog.Info("🪺 Beets integration enabled", "ip", beetsIP, "port", beetsPort)
	} else {
		slog.Info("Beets IP not configured. Beets integration will be disabled.")
	}

	// --- Validate USER_TOKENS ---

	userTokens := os.Getenv("USER_TOKENS")
	if userTokens == "" {
		slog.Error("USER_TOKENS environment variable is not set. This is required for ListenBrainz authentication.")
		os.Exit(1)
	}

	if !strings.Contains(userTokens, ":") {
		slog.Error("USER_TOKENS environment variable is malformed. It must contain at least one 'user:token' pair separated by ':'.",
			"value", userTokens,
		)
		os.Exit(1)
	}

	slog.Info("🗝️ USER_TOKENS unlocked — format looks good")

	// --- Setup HTTP Server ---
	sm := http.NewServeMux()
	sm.HandleFunc("/plex", plex.HandleRequest)
	sm.HandleFunc("/tautulli", tautulli.HandleRequest)
	sm.HandleFunc("/emby", emby.HandleRequest)
	sm.HandleFunc("/jellyfin", jellyfin.HandleRequest)
	sm.HandleFunc("/mpris", mpris.HandleRequest) // ✅ New MPRIS route
	sm.HandleFunc("/mpd", mpd.HandleRequest)     // ✅ New MPD route

	// Create the HTTP server with timeouts
	server := &http.Server{
		Addr:         address,
		Handler:      sm,
		ReadTimeout:  10 * time.Second,
		WriteTimeout: 10 * time.Second,
		IdleTimeout:  30 * time.Second,
	}

	l, err := net.Listen("tcp4", address)
	if err != nil {
		slog.Error("Failed to listen on address.",
			"address", address,
			"error", err,
		)
		os.Exit(1)
	}

	slog.Info("🧠🚀🎚️🪺🗝️🌐 Beetbrainz online and ready to scrobble")

	// Start the server in a goroutine
	go func() {
		if err := server.Serve(l); err != nil && err != http.ErrServerClosed {
			slog.Error("Server encountered an error", "error", err)
			os.Exit(1)
		}
	}()

	// 🛑 Wait for interrupt signal to gracefully shut down
	quit := make(chan os.Signal, 1)
	signal.Notify(quit, syscall.SIGINT, syscall.SIGTERM)
	sig := <-quit

	slog.Info("🛑 Signal received — shutting down Beetbrainz", "signal", sig.String())

	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()

	if err := server.Shutdown(ctx); err != nil {
		slog.Error("💥 Server forced to shutdown", "error", err)
	} else {
		slog.Info("🧹 Server shutdown complete")
	}

}
